/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/store/tlock/TLock.java,v 1.1.2.1 2004/02/05 16:05:13 mholz Exp $
 * $Revision: 1.1.2.1 $
 * $Date: 2004/02/05 16:05:13 $
 *
 * ====================================================================
 *
 * Copyright 1999-2002 The Apache Software Foundation 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package org.apache.slide.store.tlock;

import org.apache.slide.store.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Transient locking which persists until commit/rollback. ResourceLock is a shared lock (aka
 * 'ReadWriteLock'), it permits any number of readers as long as there is no writer.
 * Writer have exclusive access. Acquired before an aspect of a given ResourceId is accessed.
 * Released once during commit or abort. ResourceLock is a helper class for LockManager - nobody
 * else should know about it.
 *
 * @author    michael.hartmeier@softwareag.com
 * @author    peter.nevermann@softwareag.com
 * @version   $Revision: 1.1.2.1 $
 */
public class TLock {
    // lock types
    // Caution: the lock type is relative to the current thread. E.g. if thread A holds a
    // write Lock W, W's lock type releated to thread B is NO_LOCK.
    public static final int NO_LOCK = 0;
    public static final int READ_LOCK = 1;
    public static final int WRITE_LOCK = 2;

    //--
    
    private final long timeout;
    private final Object sync;
    private final ResourceId resourceId;
    private int waitingWriter;
    
    /** List of threads. Never contains activeWriter **/
    private final List activeReader;
    
    private Thread activeWriter;
    
    public TLock(Object sync, long timeout, ResourceId resourceId) {
        if (resourceId == null) {
            throw new IllegalArgumentException();
        }
        this.timeout = timeout;
        this.sync = sync;
        this.resourceId = resourceId;
        
        this.waitingWriter = 0;
        this.activeReader = new ArrayList();
        this.activeWriter = null;
    }
    
    public Thread getActiveWriter() {
        return activeWriter;
    }
    
    public ResourceId getResourceId() {
        return resourceId;
    }
    
    /**
     ** Obtains read or write lock for the current thread.
     ** The lock will persist til end of transaction.
     **
     ** @return current thread's lock type previously hold by current thread
     **/
    public int acquire(int type) throws TLockedException {
        switch (type) {
            case NO_LOCK:
                return getType();
            case READ_LOCK:
                return acquireReader();
            case WRITE_LOCK:
                return acquireWriter();
            default:
                throw new IllegalArgumentException("acquire failed." + type + " is invalid");
        }
    }
    
    public int getType() {
        return getType(Thread.currentThread());
    }
    
    public int getType(Thread current) {
        if (activeWriter == current) {
            return WRITE_LOCK;
        }
        if (activeReader.contains(current)) {
            return READ_LOCK;
        }
        return NO_LOCK;
    }
    
    /**
     ** @return current thread's lock type that has been released by this method.
     **/
    public int release() {
        Thread current;
        
        current = Thread.currentThread();
        if (current == activeWriter) {
            activeWriter = null;
            logSuccess("-w");
            sync.notifyAll();
            return WRITE_LOCK;
        }
        if (activeReader.remove(current)) {
            logSuccess("-r");
            sync.notifyAll();
            return READ_LOCK;
        }
        return NO_LOCK;
    }
    
    /**
     ** @return true if there is any thread with lock type != NO_LOCK
     **/
    public boolean isLocked() {
        return activeWriter != null || activeReader.size() > 0;
    }
    
    public String toString() {
        String activeWriterName = (activeWriter!=null)?activeWriter.getName():"NULL";
        return "TLock[uuri=" + resourceId+", hashCode=" + Integer.toHexString(hashCode())
            + ", writer=" + activeWriterName + ", reader=" + activeReader + ", waiting=" + waitingWriter
            + "]";
    }
    
    //--
    
    /**
     ** Obtains a read lock for the current thread.
     ** The lock will persist till end of transaction.
     **
     ** @return lock type previously hold by current thread
     **/
    private int acquireReader() throws TLockedException {
        Thread current;
        
        current = Thread.currentThread();
        if (current == activeWriter) {
            //logSuccess("acquireReader: already write locked");
            return WRITE_LOCK;
        }
        if (activeReader.contains(current)) {
            //logSuccess("acquireReader: already read locked");
            return READ_LOCK;
        }
        awaitReader();
        activeReader.add(current);
        logSuccess("+r");
        return NO_LOCK;
    }
    
    /**
     ** Obtains a write lock for the current thread.
     ** The lock will persist til end of transaction.
     **
     ** @return lock type previously hold by current thread
     **/
    private int acquireWriter() throws TLockedException {
        Thread current;
        
        current = Thread.currentThread();
        if (current == activeWriter) {
            //logSuccess("acquireWriter: already write locked");
            return WRITE_LOCK;
        }
        waitingWriter++;
        try {
            awaitWriter();
        } finally {
            waitingWriter--;
        }
        activeWriter = current;
        if (activeReader.remove(current)) {
            logSuccess("!w");
            return READ_LOCK;
        } else {
            logSuccess("+w");
            return NO_LOCK;
        }
    }
    
    private void awaitReader() throws TLockedException {
        long started;
        long remaining;
        
        if (allowReader()) {
            return;
        }
        started = System.currentTimeMillis();
        for (remaining = timeout; remaining > 0; remaining -= System.currentTimeMillis() - started) {
            logSuccess("?r");
            try {
                sync.wait(remaining);
            } catch (InterruptedException e) {
                // fall through
            }
            if (allowReader()) {
                return;
            }
        }
        throw timedOut();
    }
    
    private void awaitWriter() throws TLockedException {
        long started;
        long remaining;
        
        if (allowWriter()) {
            return;
        }
        started = System.currentTimeMillis();
        for (remaining = timeout; remaining > 0; remaining -= System.currentTimeMillis() - started) {
            logSuccess("?w");
            try {
                sync.wait(remaining);
            } catch (InterruptedException e) {
                // fall through
            }
            if (allowWriter()) {
                return;
            }
        }
        throw timedOut();
    }
    
    private boolean allowReader() {
        return activeWriter == null && waitingWriter == 0;
    }
    
    private boolean allowWriter() {
        if (activeWriter != null) {
            return false;
        }
        switch (activeReader.size()) {
            case 0:
                return true;
            case 1:
                return activeReader.get(0) == Thread.currentThread();
            default:
                return false;
        }
    }
    
    private TLockedException timedOut() {
        logSuccess("##");
        return new TLockedException("ResourceLock wait timed out: " + toString());
    }
    
    private void logSuccess(String operation) {
        //        printlnSuccess(operation);
    }
    
    private void printlnSuccess(String operation) {
        String activeWriterName = (activeWriter != null)? activeWriter.getName() : "NULL";
        List activeReaderList = new ArrayList(activeReader){
            public String toString(){
                String result = "[";
                Iterator iter = this.iterator();
                while (iter.hasNext()) {
                    result = result + ((Thread)iter.next()).getName() + ",";
                }
                return result + "]";
            }};
        
        String uuri = resourceId.getUuri();
        
        //        try { uri = XNodeRevisionDescriptor.getOneUri((IDescriptorsHandler)sync, readerDesc.getUuri()); }
        //        catch (Throwable e) { }
        
        if (operation.equals("##")) {
            
            System.out.println("   *** " +
                                   Thread.currentThread().getName() + " \t" +
                                   operation + "   " +
                                   "uuri=" + uuri
                                   + ", writer=" + activeWriterName + ", reader=" + activeReaderList + ", waiting=" + waitingWriter
                              );
            //        if (operation.equals("##")) {
            // TODO: do some more reporting [in the future]
        }
    }
}

